@@ -1,12 +1,14 @@
 from datetime import datetime
 import pandas
 import re, inspect
-import os
 import json
 
-import config
-from database import DatabaseAdapter, DatabaseConnections
-from tools import soft_del
+import jardin.config as config
+from jardin.database import DatabaseAdapter, DatabaseConnections
+from jardin.tools import soft_del, classorinstancemethod, stack_marker
+from jardin.transaction import Transaction
+from jardin.query import query
+
 
 class Collection(pandas.DataFrame):
     """
@@ -19,15 +21,6 @@ class Collection(pandas.DataFrame):
             instance.model_class = self.model_class
         return instance
 
-    @classmethod
-    def from_records(self, *args, **kwargs):
-        model_class = kwargs['model_class']
-        del kwargs['model_class']
-        collection = super(Collection, self).from_records(*args, **kwargs)
-        collection.model_class = model_class
-        return collection
-
-
     def records(self):
         """
         Returns an iterator to loop over the rows, each being an instance of the model's record class, i.e. :doc:`jardin_record` by default.
@@ -70,10 +63,9 @@ class Model(object):
     """
       Base class from which your models should inherit.
     """
-    ROLES = {"replica": "read", "master": "write"}  # to have backward compatibility
     table_name = None
     table_alias = None
-    db_names = {} # {'master': database_master_url, 'replica': database_replica_url}
+    db_names = {}
     has_many = []
     belongs_to = {}
     scopes = {}
@@ -85,10 +77,13 @@ class Model(object):
         self.attributes = dict()
         table_schema = self.__class__.table_schema()
         self.attributes[self.primary_key] = kwargs.get(self.primary_key, None)
-        for column in set(table_schema.keys() + kwargs.keys()):
+        for column in set(list(table_schema.keys()) + list(kwargs.keys())):
             self.attributes[column] = kwargs.get(
                 column,
                 table_schema.get(column, {}).get('default'))
+            # MySQL filth
+            if self.attributes[column] == '0000-00-00 00:00:00':
+                self.attributes[column] = None
             if column in kwargs:
                 del kwargs[column]
         self.init_relationships()
@@ -113,10 +108,10 @@ class Model(object):
 
     def __repr__(self):
         attrs = []
-        if 'id' in self.attributes:
-            attrs += ['id=%s' % self.id]
-        for att_name, attr_value in self.attributes.iteritems():
-            if att_name == 'id': continue
+        if self.primary_key in self.attributes:
+            attrs += ['%s=%s' % (self.primary_key, self.id)]
+        for (att_name, attr_value) in self.attributes.items():
+            if att_name == self.primary_key: continue
             attrs += ['%s=%s' % (att_name, attr_value.__repr__())]
         return '%s(%s)' % (self.__class__.__name__, ', '.join(attrs))
 
@@ -153,7 +148,8 @@ class Model(object):
         if self.persisted:
             return self.__class__.update(
                 values=self,
-                where=self.where_self)
+                where=self.where_self
+                )
         else:
             self.attributes = self.__class__.insert(values=self).attributes
 
@@ -180,14 +176,21 @@ class Model(object):
         else:
             raise RecordNotPersisted("Record's primary key is None")
 
+    def reload(self):
+        self.__init__(
+            **self.__class__.find(
+                self.attributes[self.primary_key]
+                ).attributes
+            )
+
     @property
     def persisted(self):
         return self.attributes.get(self.primary_key) is not None
 
     def init_relationships(self):
-        this_table_name = self.model_metadata()['table_name']
+        this_table_name = self._table_name()
         for h in self.has_many:
-            other_table_name = h.model_metadata()['table_name']
+            other_table_name = h._table_name()
             def func(slf, **kwargs):
                 where = kwargs.get('where', {})
                 where.update(**{
@@ -198,29 +201,17 @@ class Model(object):
             setattr(self.__class__, other_table_name, func)
 
     @classmethod
-    def collection(self, *args, **kwargs):
-        collection = self.collection_class(*args, **kwargs)
+    def collection_instance(self, result=None):
+        if isinstance(result, list) and len(result) and isinstance(result[0], dict):
+            collection = self.collection_class.from_records(result)
+        else:
+            collection = self.collection_class(result)
         collection.model_class = self
         return collection
 
-    @classmethod
-    def collection_instance(self, result):
-        return self.collection_class.from_records(
-            result[0],
-            columns=result[1],
-            coerce_float=True,
-            model_class=self
-            )
-
     @classmethod
     def stack_mark(self, stack, db_conn=None):
-        filename = stack[1][1]
-        function_name = stack[1][3]
-        line_number = stack[1][2]
-        stack = [filename, function_name, str(line_number)]
-        if db_conn:
-          stack = [db_conn.connection().dsn] + stack
-        return ':'.join(stack)
+        return stack_marker(stack, db_conn=db_conn)
 
     @classmethod
     @soft_del
@@ -254,7 +245,7 @@ class Model(object):
         """
         db_adapter = self.db_adapter(
             db_name=kwargs.get('db'),
-            role=kwargs.get('role') or 'replica'
+            role=kwargs.get('role', 'replica')
             )
 
         kwargs['stack'] = self.stack_mark(
@@ -278,27 +269,17 @@ class Model(object):
         :type role: string
         :returns: ``jardin.Collection`` collection, which is a ``pandas.DataFrame``.
         """
-        kwargs['stack'] = self.stack_mark(inspect.stack())
-        
-        if filename:
-            filename = os.path.join(os.environ['PWD'], filename)
-        
-        kwargs['where'] = kwargs.get('where', kwargs.get('params'))
-        
-        results = self.db_adapter(
-            db_name=kwargs.get('db'),
-            role=kwargs.get('role') or 'replica'
-            ).raw_query(
-                sql=sql,
-                filename=filename,
-                **kwargs
-                )
+        results = query(
+            sql=sql,
+            filename=filename,
+            db=self.db_names[kwargs.get('role', 'replica')],
+            **kwargs
+            )
 
-        if results:
-            return self.collection_instance(results)
-        else:
+        if results is None:
             return None
-
+        else:
+            return self.collection_instance(results)
 
     @classmethod
     @soft_del
@@ -317,14 +298,15 @@ class Model(object):
         :returns: integer
         """
         if 'select' in kwargs:
-            kwargs['select'] = 'COUNT(%s)' % kwargs['select']
+            kwargs['select'] = {'cnt': 'COUNT(%s)' % kwargs['select']}
         else:
-            kwargs['select'] = 'COUNT(*)'
+            kwargs['select'] = {'cnt': 'COUNT(*)'}
 
-        return self.db_adapter(
+        res = self.db_adapter(
             db_name=kwargs.get('db'),
-            role=kwargs.get('role') or 'replica'
-            ).select(**kwargs)[0][0]['count']
+            role=kwargs.get('role', 'replica')
+            ).select(**kwargs)
+        return res.cnt[0]
 
     @classmethod
     def insert(self, **kwargs):
@@ -341,23 +323,27 @@ class Model(object):
         if isinstance(values, self):
             values = values.attributes.copy()
         if isinstance(values, dict):
-            for k, v in values.iteritems():
+            for (k, v) in list(values.items()):
                 if v is None:
                     del kwargs['values'][k]
+            
         kwargs['stack'] = self.stack_mark(inspect.stack())
         kwargs['primary_key'] = self.primary_key
+
         column_names = self.table_schema().keys()
         now = datetime.utcnow()
-        for field in ('created_at', 'updated_at'):
-            if field in column_names:
+        for field in [f for f in ('created_at', 'updated_at') if f in column_names]:
+            if isinstance(kwargs['values'], dict):
+                kwargs['values'][field] = kwargs['values'].get(field, now)
+            else:
                 kwargs['values'][field] = now
         results = self.db_adapter(role='master').insert(**kwargs)
         return self.record_or_model(results)
 
     @classmethod
     def record_or_model(self, results):
-        if len(results[0]) == 1:
-            return self(**results[0][0])
+        if len(results) == 1:
+            return self(**results.to_dict(orient='records')[0])
         else:
             return self.collection_instance(results)
 
@@ -377,11 +363,21 @@ class Model(object):
         column_names = self.table_schema().keys()
         now = datetime.utcnow()
         if 'updated_at' in column_names:
-            kwargs['values']['updated_at'] = now
-
+            if 'updated_at' not in kwargs['values']:
+                kwargs['values']['updated_at'] = now
         results = self.db_adapter(role='master').update(**kwargs)
         return self.record_or_model(results)
 
+    @classorinstancemethod
+    def touch(self, **kwargs):
+        if type(self) == type:
+            kwargs['values'] = {'updated_at': datetime.utcnow()}
+            return self.update(**kwargs)
+        else:
+            self.__class__.touch(where=self.where_self)
+            self.reload()
+
+
     @classmethod
     @soft_del
     def delete(self, **kwargs):
@@ -403,7 +399,7 @@ class Model(object):
         return self.collection_instance(
             self.db_adapter(
                 db_name=kwargs.get('db'),
-                role=kwargs.get('role') or 'replica'
+                role=kwargs.get('role', 'replica')
                 ).select(
                     where='created_at IS NOT NULL',
                     order='created_at DESC',
@@ -421,13 +417,12 @@ class Model(object):
         :returns: an instance of the model.
         """
         try:
-            return self(**self.db_adapter(
-                db_name=kwargs.get('db'),
-                role=kwargs.get('role') or 'replica'
-                ).select(
+            return self(
+                **self.select(
                     where=values,
-                    limit=1
-                    )[0][0]
+                    limit=1,
+                    **kwargs
+                    ).to_dict(orient='records')[0]
                 )
         except IndexError:
             return None
@@ -438,60 +433,65 @@ class Model(object):
         Finds a record by its id in the model's table in the replica database.
         :returns: an instance of the model.
         """
-        return self.find_by(values={'id': id}, **kwargs)
+        return self.find_by(values={self.primary_key: id}, **kwargs)
 
     @classmethod
     def db_adapter(self, role='replica', db_name=None):
-        if not hasattr(self, '_db_adapter'):
-            self._db_adapter = {}
-        
-        if role not in self._db_adapter:
-          self._db_adapter[role] = DatabaseAdapter(
-              self.db(
-                  role=role,
-                  db_name=db_name
-                  ), 
-              self.model_metadata()
-              )
-
-        return self._db_adapter[role]
+        if not hasattr(self, '_db_metadata'):
+            self._db_metadata = {}
+        db_name = db_name or self.db_names.get(role)
+        key = '%s_%s' % (db_name, role)
+        if key not in self._db_metadata:
+            self._db_metadata[key] = self.model_metadata()
+        return DatabaseAdapter(
+            self.db(role=role, db_name=db_name),
+            self._db_metadata[key]
+            )
 
     @classmethod
-    def model_metadata(self):
-        table_name = self.table_name or self._default_table_name()
-        table_alias = self.table_alias or self._default_table_alias(table_name)
-        return {
-            'table_name': table_name,
-            'table_alias': table_alias,
+    def model_metadata(self, include_schema=True):
+        metadata = {
+            'table_name': self._table_name(),
+            'table_alias': self._table_alias(),
             'belongs_to': self.belongs_to,
             'scopes': self.scopes
         }
+        if include_schema:
+            metadata['table_schema'] = self.table_schema()
+        return metadata
 
     @staticmethod
     def _default_table_alias(table_name):
         return ''.join([w[0] for w in table_name.split('_')])
 
+    @classmethod
+    def _table_name(self):
+        if self.__dict__.get('__table_name') is None:
+            self.__table_name = self.table_name or self._default_table_name()
+        return self.__table_name
+
+    @classmethod
+    def _table_alias(self):
+        if self.__dict__.get('__table_alias') is None:
+            self.__table_alias = self.table_alias or self._default_table_alias(
+                self._table_name()
+                )
+        return self.__table_alias
+
     @classmethod
     def _default_table_name(self):
         import inflect
         name = self.__name__
-        #s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
         s1 = re.sub('([A-Z])', r'_\1', name)[1:]
         s1 = s1.split('_')
+        s1 = list(map(lambda x: x.lower(), s1))
         s1[-1] = inflect.engine().plural_noun(s1[-1])
-        return '_'.join(map(lambda x: x.lower(), s1))
+        return '_'.join(s1)
 
     @classmethod
-    def db(self, role='replica', db_name = None):
-        if not hasattr(self, '_db'): self._db = {}
-        name = self.db_names.get(
-            role,
-            self.db_names.get(self.ROLES.get(role))) if db_name is None else db_name
-        
-        if name not in self._db:
-            self._db[name] = DatabaseConnections.connection(name)
-
-        return self._db[name]
+    def db(self, role='replica', db_name=None):
+        name = db_name or self.db_names.get(role)
+        return DatabaseConnections.connection(name)
 
     @classmethod
     def _use_replica(self, **kwargs):
@@ -530,7 +530,7 @@ class Model(object):
         """
         Enables multiple statements to be ran within a single transaction, see :doc:`features`.
         """
-        return Transaction(self)
+        return Transaction(self.db(role='master').name)
 
     @classmethod
     def table_schema(self):
@@ -540,33 +540,38 @@ class Model(object):
         :returns: dict
         """
         if self.__dict__.get('_table_schema') is None:
-            self._table_schema = {}
+            self._table_schema = None
+            table_schema = {}
             for row in self.query_schema():
-                default = row['column_default']
-                name = row['column_name']
+                name, default, dtype = self.db().lexicon.column_info(row)
                 if isinstance(default, str):
                     json_matches = re.findall(r"^\'(.*)\'::jsonb$", default)
                     if len(json_matches) > 0:
                         default = json.loads(json_matches[0])
                 if name == self.primary_key:
                     default = None
-                self._table_schema[name] = {'default': default}
-
+                table_schema[name] = {'default': default, 'type': dtype}
+            if len(table_schema):
+                self._table_schema = table_schema
         return self._table_schema
 
     @classmethod
     def query_schema(self):
-        table_name = self.model_metadata()['table_name']
-        return self.db_adapter().raw_query(
-            sql="SELECT column_name, column_default FROM " \
-            "information_schema.columns WHERE " \
-            "table_name=%(table_name)s AND table_schema='public';",
-            where={'table_name': table_name}
-            )[0]
+        db_adapter = DatabaseAdapter(
+            self.db(role='replica', db_name=self.db_names.get('replica')),
+            self.model_metadata(include_schema=False)
+            )
+        return db_adapter.raw_query(
+            sql=db_adapter.db.lexicon.table_schema_query(self._table_name()),
+            where={'table_name': self._table_name()}
+            ).to_dict(orient='records')
 
     @classmethod
     def clear_caches(self):
         self._table_schema = None
+        self._db_metadata = {}
+        self.__table_alias = None
+        self.__table_name = None
 
 class ModelIterator(object):
 
@@ -586,21 +591,3 @@ class ModelIterator(object):
             return record
         else:
             raise StopIteration()
-
-
-class Transaction(object):
-
-    def __init__(self, model):
-        self._model = model
-        self._connection = self._model.db(role='master')
-
-    def __enter__(self):
-        self._connection.autocommit = False
-        self._model.query(sql='BEGIN;', role='master')
-
-    def __exit__(self, exc_type, exc_value, traceback):
-        if exc_type is None:
-            self._model.query(sql='COMMIT;', role='master')
-        else:
-            self._model.query(sql='ROLLBACK;', role='master')
-        self._connection.autocommit = True
